# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import enum
import itertools as it
from typing import Iterator, List, NamedTuple


class Chunk(NamedTuple):
    new_lineno: int
    new_offset: int
    del_lineno: int
    del_offset: int

    def newlines(self) -> range:
        return range(self.new_lineno, self.new_lineno + self.new_offset)

    def dellines(self) -> range:
        return range(self.del_lineno, self.del_lineno + self.del_offset)


class ChangeKind(enum.Enum):
    ADDED = "A"
    MODIFIED = "M"
    DELETED = "D"


class Change(NamedTuple):
    filename: str
    kind: ChangeKind
    chunks: List[Chunk]

    def has_newlines(self) -> bool:
        return any(c.new_offset > 0 for c in self.chunks)

    def has_dellines(self) -> bool:
        return any(c.del_offset > 0 for c in self.chunks)

    def newlines(self) -> Iterator[int]:
        return it.chain.from_iterable((c.newlines() for c in self.chunks))

    def dellines(self) -> Iterator[int]:
        return it.chain.from_iterable((c.dellines() for c in self.chunks))


class Commit(NamedTuple):
    hash: str
    changes: List[Change]



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import argparse
import sys
from typing import List, Optional

from gitchurn import gitchurn


def main(argv: Optional[List[str]] = None) -> int:
    if argv is None:
        argv = sys.argv[1:]

    parser = argparse.ArgumentParser(
        prog="gitchurn", description="calculate change churn below the file-level"
    )
    parser.add_argument(
        "--git-repo",
        dest="git_repo",
        default=".",
        help="path to the git repository (default: current directory)",
    )
    parser.add_argument(
        "--git-path",
        dest="git_bin",
        default="git",
        help="path to git binary (default: git)",
    )
    parser.add_argument(
        "--ctags-path",
        dest="ctags_bin",
        default="ctags",
        help="path to the universal-ctags binary (default: ctags)",
    )
    parser.add_argument(
        "--git-log-args",
        dest="git_log_args",
        default="",
        help="any additional arguments to pass to git-log (optional)",
    )
    parser.add_argument(
        "--max-changes",
        dest="max_changes",
        help="exclude commits that changed more than this number of files (optional)",
    )

    formatter_factory = gitchurn.TagFormatterFactory()

    parser.add_argument(
        "--tag-format",
        choices=formatter_factory.kinds(),
        default="human",
        dest="tag_format",
        help="how to display the tags (default: human)",
    )

    args = parser.parse_args(argv)
    git_driver = gitchurn.GitDriver(
        git_bin=args.git_bin, git_repo=args.git_repo, git_log_args=args.git_log_args
    )
    ctags_driver = gitchurn.CTagsDriver(ctags_bin=args.ctags_bin)
    tag_provider = gitchurn.TagProvider(git_driver, ctags_driver)
    churn_provider = gitchurn.ChurnProvider(tag_provider)
    record_provider = gitchurn.LogRecordProvider(churn_provider, git_driver)

    max_changes = int(args.max_changes) if args.max_changes else None
    formatter = formatter_factory.create(args.tag_format)
    for record in record_provider.fetch(formatter, max_changes):
        print("{}\t{}\t{}".format(record.commit, record.churn, record.tag))
    return 0


if __name__ == "__main__":
    sys.exit(main())



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import abc
import json
import shlex
import subprocess as sp
from functools import reduce
from itertools import chain
from typing import (
    Counter,
    FrozenSet,
    Iterator,
    List,
    Mapping,
    NamedTuple,
    Optional,
    Tuple,
)

from gitchurn import gitparser, ir

ENCODING = "UTF-8"

Tag = Mapping[str, str]
CanonTag = FrozenSet[Tuple[str, str]]


def has_lineno(tag: Tag, lineno: int) -> bool:
    line = tag.get("line")
    end = tag.get("end")
    if line is None:
        raise RuntimeError("Tag produced by ctags is missing `line` property.")
    # Sometimes `end` is missing. This seems like a bug with universal-ctags.
    # It seems this only happens for the last tag in the file.
    return lineno >= int(line) and (end is None or lineno <= int(end))


def count_linenos(tag: Tag, linenos: Iterator[int]) -> int:
    return sum(1 for i in linenos if has_lineno(tag, i))


def to_canon(tag: Tag) -> CanonTag:
    return frozenset(
        (k, v) for k, v in tag.items() if k not in ["line", "end", "_type"]
    )


def run(args: List[str]) -> str:
    return sp.run(args, encoding=ENCODING, capture_output=True, check=True).stdout


class GitDriver:
    def __init__(self, **kwargs: str) -> None:
        git_bin = kwargs.get("git_bin", "git")
        git_repo = kwargs.get("git_repo", ".")
        self._init_args = [git_bin, "-C", git_repo]
        self._log_args = shlex.split(kwargs.get("git_log_args", ""))

    def show(self, filename: str, hash: str) -> str:
        args = self._init_args + ["show", "{}:{}".format(hash, filename)]
        return run(args)

    def files(self, ref: str) -> List[str]:
        args = self._init_args + ["ls-tree", "-r", "--name-only", ref]
        return run(args).splitlines()

    def log(self) -> Iterator[str]:
        args = self._init_args + ["log"] + gitparser.GIT_LOG_ARGS + self._log_args
        proc = sp.Popen(args, encoding=ENCODING, stdout=sp.PIPE)
        assert proc.stdout is not None
        return proc.stdout


class CTagsDriver:
    def __init__(self, **kwargs: str) -> None:
        self._init_args = [kwargs.get("ctags_bin", "ctags")]

    def generate_tags(self, filename: str, text: str) -> List[Tag]:
        # TODO: Allow customizing these arguments from the command line
        args = self._init_args + ["--_interactive", "--fields=FzZNpen", "--extras=+f"]
        proc = sp.Popen(args, encoding=ENCODING, stdout=sp.PIPE, stdin=sp.PIPE)
        cin = '{{"command":"generate-tags", "filename":"{}", "size":{}}}\n{}'.format(
            filename, len(text.encode(ENCODING)), text
        )
        # We skip the first and last message because they are not tags.
        lines = proc.communicate(cin)[0].splitlines()[1:-1]
        return [json.loads(line) for line in lines]


class TagProvider:
    def __init__(self, git: GitDriver, ctags: CTagsDriver):
        self._git = git
        self._ctags = ctags

    def get_tags(self, filename: str, hash: str) -> List[Tag]:
        return self._ctags.generate_tags(filename, self._git.show(filename, hash))

    def get_parent_tags(self, filename: str, hash: str) -> List[Tag]:
        return self.get_tags(filename, "{}^".format(hash))


class ChurnProvider:
    def __init__(self, tag_provider: TagProvider):
        self._tag_provider = tag_provider

    def get_churn(self, commit: ir.Commit) -> Counter[CanonTag]:
        adds = (self.get_adds(commit.hash, c) for c in commit.changes)
        dels = (self.get_dels(commit.hash, c) for c in commit.changes)
        return reduce(lambda a, b: a + b, chain(adds, dels), Counter())

    def get_adds(self, hash: str, change: ir.Change) -> Counter[CanonTag]:
        count: Counter[CanonTag] = Counter()
        if not change.has_newlines():
            return count
        for tag in self._tag_provider.get_tags(change.filename, hash):
            count[to_canon(tag)] = count_linenos(tag, change.newlines())
        return count

    def get_dels(self, hash: str, change: ir.Change) -> Counter[CanonTag]:
        count: Counter[CanonTag] = Counter()
        if not change.has_dellines():
            return count
        for tag in self._tag_provider.get_parent_tags(change.filename, hash):
            count[to_canon(tag)] = count_linenos(tag, change.dellines())
        return count


class TagFormatter(abc.ABC):
    @abc.abstractmethod
    def format(self, tag: CanonTag) -> str:
        pass


class HumanFormatter(TagFormatter):
    def format(self, tag: CanonTag) -> str:
        d = {k: v for k, v in tag}
        return "{} > {} ({})".format(d["path"], d["name"], d["kind"])


class JsonFormatter(TagFormatter):
    def format(self, tag: CanonTag) -> str:
        return json.dumps({k: v for k, v in tag}, sort_keys=True)


class TagFormatterFactory:
    def __init__(self) -> None:
        self._kinds = ["human", "json"]

    def kinds(self) -> List[str]:
        return self._kinds

    def create(self, kind: str) -> TagFormatter:
        if kind not in self.kinds():
            raise ValueError("Invalid `kind`.")
        if kind == "human":
            return HumanFormatter()
        return JsonFormatter()


class LogRecord(NamedTuple):
    commit: str
    churn: int
    tag: str


class LogRecordProvider:
    def __init__(self, churn_provider: ChurnProvider, git_driver: GitDriver):
        self._churn = churn_provider
        self._git = git_driver

    def fetch(
        self, formatter: TagFormatter, max_changes: Optional[int] = None
    ) -> Iterator[LogRecord]:
        for commit in gitparser.parse(self._git.log()):
            # Exclude commits that changed too many files
            if (max_changes is not None) and (len(commit.changes) > max_changes):
                continue
            for tag, churn in self._churn.get_churn(commit).items():
                yield LogRecord(commit.hash, churn, formatter.format(tag))



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import re
from typing import Iterator, List, Optional

from gitchurn import ir

STR_COMMIT = "commit "
STR_DIFF = "diff --git"
STR_NEW_FILE = "new file mode"
STR_DEL_FILE = "deleted file mode"
STR_FROM = "--- a/"
STR_TO = "+++ b/"
STR_CHUNK = "@@ -"

RE_CHUNK = re.compile(r"^@@ -(\d+)(?:,(\d+))? \+(\d+)(?:,(\d+))? @@")

GIT_LOG_ARGS = [
    # Format commits so the output log can be parsed.
    "--pretty=short",
    "--no-decorate",
    # Disable rename and move detection.
    "--no-renames",
    "--no-color-moved",
    # Display patch information with each commit and show only changed lines.
    "--unified=0",
    # Use the histogram algorithm for diffs. This is best for source code.
    # https://link.springer.com/article/10.1007/s10664-019-09772-z
    "--histogram",
    # Only show Added (A), Deleted (D), and Modified (M) files.
    "--diff-filter=ADM",
    # Hide merge commits.
    "--no-merges"
]


class ChangeBuilder:
    def __init__(self) -> None:
        self.reset()

    def reset(self) -> None:
        self._filename: Optional[str] = None
        self._kind: ir.ChangeKind = ir.ChangeKind.MODIFIED
        self._chunks: List[ir.Chunk] = []

    def set_filename(self, filename: str) -> None:
        self._filename = filename

    def set_kind(self, kind: ir.ChangeKind) -> None:
        self._kind = kind

    def add_chunk(self, chunk: ir.Chunk) -> None:
        self._chunks.append(chunk)

    def is_valid(self) -> bool:
        return self._filename is not None

    def finalize(self) -> ir.Change:
        if self._filename is None:
            raise RuntimeError("Cannot finalize change without a filename.")
        change = ir.Change(self._filename, self._kind, self._chunks)
        self.reset()
        return change


class CommitBuilder:
    def __init__(self) -> None:
        self.reset()

    def reset(self) -> None:
        self._hash: Optional[str] = None
        self._changes: List[ir.Change] = []

    def set_hash(self, hash: str) -> None:
        self._hash = hash

    def add_change(self, change: ir.Change) -> None:
        self._changes.append(change)

    def is_valid(self) -> bool:
        return self._hash is not None

    def finalize(self) -> ir.Commit:
        if self._hash is None:
            raise RuntimeError("Cannot finalize commit without a hash.")
        commit = ir.Commit(self._hash, self._changes)
        self.reset()
        return commit


def parse_chunk(text: str) -> ir.Chunk:
    match = RE_CHUNK.match(text)
    if match is None:
        raise RuntimeError("Invalid chunk header: {}".format(text))
    a, b, c, d = match.groups()
    del_offset = int(b) if b else 1
    new_offset = int(d) if d else 1
    return ir.Chunk(int(c), new_offset, int(a), del_offset)


def parse(lines: Iterator[str]) -> Iterator[ir.Commit]:
    # These builders hold our intermediate state.
    commit_builder = CommitBuilder()
    change_builder = ChangeBuilder()

    # Iterate line by line through the log.
    for line in (l.rstrip() for l in lines):
        if line.startswith(STR_COMMIT):
            # Yield the previous commit before starting a new one.
            if change_builder.is_valid():
                commit_builder.add_change(change_builder.finalize())
            if commit_builder.is_valid():
                yield commit_builder.finalize()
            # Set the hash.
            commit_builder.set_hash(line[len(STR_COMMIT) :])
        # Finalize the current change before starting a new one.
        elif line.startswith(STR_DIFF) and change_builder.is_valid():
            commit_builder.add_change(change_builder.finalize())
        # Below here we just update our intermediate state.
        elif line.startswith(STR_FROM):
            change_builder.set_filename(line[len(STR_FROM) :])
        elif line.startswith(STR_TO):
            change_builder.set_filename(line[len(STR_TO) :])
        elif line.startswith(STR_NEW_FILE):
            change_builder.set_kind(ir.ChangeKind.ADDED)
        elif line.startswith(STR_DEL_FILE):
            change_builder.set_kind(ir.ChangeKind.DELETED)
        elif line.startswith(STR_CHUNK):
            change_builder.add_chunk(parse_chunk(line))

    # Yield the last commit. (Except when the log file is empty.)
    if change_builder.is_valid():
        commit_builder.add_change(change_builder.finalize())
    if commit_builder.is_valid():
        yield commit_builder.finalize()
